//----------------------------------------------------------------------------
/*
 * File: lcd.c
 * - functions to manage with LCD display
 * 
 * Part of Emulation of PMD 32 drive.
 * Based on MMC/SD/SDHC card library by Roland Riegel
 *   - http://www.roland-riegel.de/sd-reader/
 *
 * Author: Roman Brik { pmd85emu (at) gmail.com }
 * Web: http://pmd85.borik.net
 * Date: 15 Sep 2009
 * Version 1.3
*/
//----------------------------------------------------------------------------
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "lcd.h"
//----------------------------------------------------------------------------
static uint8_t values[16] = { 0, 32, 16, 48, 8, 40, 24, 56,
                              4, 36, 20, 52, 12, 44, 28, 60 };
//----------------------------------------------------------------------------
// local static function for sending command or data to the display
static void lcd_cmd_data(uint8_t rs, uint8_t val)
{
  if (rs)
    lcd_rs_data();
  else
    lcd_rs_cmd();
  lcd_e_on();
  LCD_PORT_DATA = ((LCD_PORT_DATA & (~LCD_DATA_MASK)) | values[((val & 0xF0) >> 4)]); // upper nibble
  _delay_us(100);
  lcd_e_off();
  _delay_us(100);
  lcd_e_on();
  LCD_PORT_DATA = ((LCD_PORT_DATA & (~LCD_DATA_MASK)) | values[val & 0x0F]); // lower nibble
  _delay_us(100);
  lcd_e_off();

  if (rs || (val & 0xFC))
    _delay_us(100); // for data or others commands
  else
    _delay_ms(1.7); // clear display or return home commands
}
//----------------------------------------------------------------------------
// local static helper function for display initialisation
static void lcd_init_aux(uint8_t val)
{
  lcd_rs_cmd();
  lcd_e_on();
  LCD_PORT_DATA = ((LCD_PORT_DATA & (~LCD_DATA_MASK)) | values[val]);

  _delay_us(100);
  lcd_e_off();
  _delay_us(100);
}
//----------------------------------------------------------------------------
// display initiaisation
void lcd_init(void)
{
  lcd_data_config();
  lcd_e_config();
  lcd_rs_config();
  lcd_backlight_config();
  lcd_backlight_off();

  _delay_ms(15);
  lcd_init_aux(0x03);
  _delay_ms(4.1);
  lcd_init_aux(0x03);
  lcd_init_aux(0x03);
  lcd_init_aux(0x02);

  lcd_backlight_on();
  lcd_cmd_data(LCD_CMD, LCD_CMD_FUNCTION);
  lcd_cmd_data(LCD_CMD, LCD_CMD_CLR);
  lcd_cmd_data(LCD_CMD, LCD_CMD_DISPLAY_ON);
  lcd_cmd_data(LCD_CMD, LCD_CMD_ENTRY_MODE);
}
//----------------------------------------------------------------------------
// turn ON display
void lcd_on(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_DISPLAY_ON);
}
//----------------------------------------------------------------------------
// turns OFF display
void lcd_off(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_DISPLAY_OFF);
}
//----------------------------------------------------------------------------
// clear display
void lcd_clr(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_CLR);
}
//----------------------------------------------------------------------------
// turn ON cursor
void lcd_cursor_on(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_CURSOR_ON);
}
//----------------------------------------------------------------------------
// turn OFF cursor
void lcd_cursor_off(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_DISPLAY_ON);
}
//----------------------------------------------------------------------------
// turn ON blinking cursor
void lcd_cursor_blink(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_CURSOR_BLINK);
}
//----------------------------------------------------------------------------
// move cursor home
void lcd_home(void)
{
  lcd_cmd_data(LCD_CMD, LCD_CMD_HOME);
}
//----------------------------------------------------------------------------
// set new print position
void lcd_set_position(uint8_t row, uint8_t col)
{
  register uint8_t a;

  // remap lines into proper order
  switch (row)
  {
    case 1 :
      a = LCD_LINE1_DDRAM_ADDR + col;
      break;

    case 2 :
      a = LCD_LINE2_DDRAM_ADDR + col;
      break;

    case 3 :
      a = LCD_LINE3_DDRAM_ADDR + col;
      break;

    case 0 :
    default :
      a = LCD_LINE0_DDRAM_ADDR + col;
      break;
  }

  // set data address
  lcd_cmd_data(LCD_CMD, LCD_CMD_DDRAM_ADDR | a);
}
//----------------------------------------------------------------------------
// shift characters on display n positions right
void lcd_shift_left(uint8_t n)
{
  for (uint8_t i = 0; i < n; i++)
  {
    lcd_cmd_data(LCD_CMD, LCD_CMD_SHIFT_LEFT);
  }
}
//----------------------------------------------------------------------------
// shift characters on display n positions right
void lcd_shift_right(uint8_t n)
{
  for (uint8_t i = 0; i < n; i++)
  {
    lcd_cmd_data(LCD_CMD, LCD_CMD_SHIFT_RIGHT);
  }
}
//----------------------------------------------------------------------------
// move cursor by n poisitions left
void lcd_cursor_left(uint8_t n)
{
  for (uint8_t i = 0; i < n; i++)
  {
    lcd_cmd_data(LCD_CMD, LCD_CMD_CURSOR_LEFT);
  }
}
//----------------------------------------------------------------------------
// move cursor by n poisitions right
void lcd_cursor_right(uint8_t n)
{
  for (uint8_t i = 0; i < n; i++)
  {
    lcd_cmd_data(LCD_CMD, LCD_CMD_CURSOR_RIGHT);
  }
}
//----------------------------------------------------------------------------
// definition of char symbol in CGRAM
/* example:
  const prog_uint8_t backslash[] =
  {
    0b00000000, // back slash
    0b00010000,
    0b00001000,
    0b00000100,
    0b00000010,
    0b00000001,
    0b00000000,
    0b00000000
  };
  lcd_define_char(backslash, 0);
*/
void lcd_define_char(const prog_uint8_t *pc, uint8_t char_code)
{
  uint8_t a, i;

  a = (char_code << 3) | LCD_CMD_CGRAM_ADDR;
  for (i = 0; i < 8; i++)
  {
    lcd_cmd_data(LCD_CMD, a++);
    lcd_cmd_data(LCD_DATA, pgm_read_byte(&pc[i]));
  }
}
//----------------------------------------------------------------------------
// display character at current position
void lcd_disp_char(char c)
{
  lcd_cmd_data(LCD_DATA, (uint8_t)c);
}
//----------------------------------------------------------------------------
// display character at requested position
void lcd_disp_char_at(char c, uint8_t row, uint8_t col)
{
  lcd_set_position(row, col);
  lcd_cmd_data(LCD_DATA, (uint8_t)c);
}
//----------------------------------------------------------------------------
// display string at current position
void lcd_disp_string(char *str)
{
  // check to make sure we have a good pointer
  if (!str)
    return;

  // print string
  while (*str)
  {
    lcd_cmd_data(LCD_DATA, (uint8_t)*str++);
  }
}
//----------------------------------------------------------------------------
// display string at requested position
void lcd_disp_string_at(char *str, uint8_t row, uint8_t col)
{
  lcd_set_position(row, col);
  lcd_disp_string(str);
}
//----------------------------------------------------------------------------
// display string from flash memory at current position
/* example:
   PGM_P hello = PSTR("Hello World");
   lcd_disp_string_p(hello);
   or
   lcd_disp_string_p(PSTR("Hello World"));
*/
void lcd_disp_string_p(PGM_P str)
{
  uint8_t i, c;

  if (!str)
    return;

  i = 0;
  while (1)
  {
    c = pgm_read_byte(&str[i++]);
    if (!c)
      return;

    lcd_cmd_data(LCD_DATA, c);
  }
}
//----------------------------------------------------------------------------
// display string from flash memory at requested position
/* example:
   PGM_P hello = PSTR("Hello World");
   lcd_disp_string_at_p(hello, 1, 2);
   or
   lcd_disp_string_at_p(PSTR("Hello World"), 1, 2);
*/
void lcd_disp_string_at_p(PGM_P str, uint8_t row, uint8_t col)
{
  lcd_set_position(row, col);
  lcd_disp_string_p(str);
}
//----------------------------------------------------------------------------
