//----------------------------------------------------------------------------
/*
 * File: pmd32.c
 * - emulation of PMD 32 drive
 *
 * Part of Emulation of PMD 32 drive.
 * Based on MMC/SD/SDHC card library by Roland Riegel
 *   - http://www.roland-riegel.de/sd-reader/
 *
 * Author: Roman Brik { pmd85emu (at) gmail.com }
 * Web: http://pmd85.borik.net
 * Date: 15 Sep 2009
 * Version 1.3
*/
//----------------------------------------------------------------------------
#define TITLE     PSTR("PMD 32-SD")
#define VERSION   PSTR("v1.3")
#define DATETIME  PSTR("5 Sep 2009")
//----------------------------------------------------------------------------
#include <inttypes.h>
#include <string.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "pgm_strings.h"
#include "sd_raw.h"
#include "partition.h"
#include "fat.h"
#include "timer.h"
#include "btn.h"
#include "lcd.h"
#include "ui.h"
#include "apputil.h"
#include "config.h"
#include "pmd32.h"

#if (1 & DEBUG)
  #include "uart.h"
  #define DEBUG_SD  1
  #define DEBUG_CMD 1
  #define DEBUG_OP  1
  #define DEBUG_CRC 1
#else
  #define DEBUG_SD  0
  #define DEBUG_CMD 0
  #define DEBUG_OP  0
  #define DEBUG_CRC 0
#endif

#define DEBUG_ON_LCD      0
#define DEBUG_ON_LCD_2    0
//----------------------------------------------------------------------------
#define EXTRA_COMMANDS    1
//----------------------------------------------------------------------------
// PMD 32 drive structure
typedef struct {
  struct fat_file_struct *hnd;      // pointer to file structure
  uint8_t wp;                       // write protect flag
} DRIVES;
//----------------------------------------------------------------------------
// global variables
static uint8_t buffer[PHYSICAL_SECTOR_SIZE + 1];  // general and sector buffer
static uint8_t command;
static uint8_t crc;
static uint8_t drvnum;
static uint8_t track;
static uint8_t sector;
static struct fat_fs_struct *fs;
static struct partition_struct* part;
static DRIVES drives[DRIVES_COUNT];
static char str[LCD_DISPLAY_COLUMNS + 1];
//----------------------------------------------------------------------------
// function prototypes
void byte_available(void);
uint8_t read_byte(uint8_t *byte, uint32_t to_ms);
uint8_t read_byte_ex(uint8_t *byte);
uint8_t read_byte_crc_check(void);
uint8_t read_sector_track(void);

uint8_t send_byte(uint8_t byte, uint32_t to_ms);
void byte_accepted(void);

uint8_t prepare_sector(void);
uint8_t write_sector(void);

void insert_disk(uint8_t drive, char *file_path, uint8_t wp);
void remove_disk(uint8_t drive);
uint8_t create_image(struct fat_dir_struct *parent_dir, const char *file_name);
void settings(void);

void close_sd_card(void);
void prepare_sd_card(void);
//----------------------------------------------------------------------------
int main(void)
{
  uint8_t val, *point;
  int8_t op;
  uint16_t address, length = 0;
#if EXTRA_COMMANDS
  char cur_dir[FAT_MAX_PATH], *p;
  struct fat_dir_struct *ds, *ds_new;
  struct fat_dir_entry_struct de;
#endif

  // we will just use ordinary idle mode
  set_sleep_mode(SLEEP_MODE_IDLE);

  // init timer
  timer_init();
  // inti LCD
  lcd_init();
  // init buttons
  btn_init();
#if DEBUG
  // init uart
  uart_init();
#endif
  // configure SD card port pins
  sd_raw_configure();

  // configure communication parallel port
  // data port to Z-state
  pmd32_data_port_Z();
  // ACK, STB and DIR pins as outputs, log. H
  // IBF and OBF pins as input with pull-ups
  pmd32_ctrl_config();
  pmd32_dir_config();

  sei();  // enable interrupt

#if DEBUG
  uart_puts_p(PSTR("PMD 32-SD Start\n"));
#endif

  // show "welcome" text
  strcpy_P((char *)buffer, TITLE);
  strcat_P((char *)buffer, PSTR(" "));
  strcat_P((char *)buffer, VERSION);
  lcd_backlight_on();
  lcd_clr();
  lcd_disp_string(apputil_center_text(str, (char *)buffer, LCD_DISPLAY_COLUMNS));
  timer_wait_secs(3);

  part = 0;
  fs = 0;
#if EXTRA_COMMANDS
  ds = 0;
  ds_new = 0;
  *cur_dir = 0;
#endif
  while (1)
  {
    // wait for inserting of card
    prepare_sd_card();

    // read configuration from EEPROM
    if (config_check_card_sn())
    {
      for (op = 0; op < DRIVES_COUNT; op++)
      {
        config_read_pmd32_img_path(op, (char *)buffer);
        config_read_pmd32_wp(op, &val);
        insert_disk(op, (char *)buffer, val);
      }
    }
    else
    {
      // another card is inserted
      // clear previous settings
      *buffer = 0;
      for (op = 0; op < DRIVES_COUNT; op++)
      {
        config_write_pmd32_img_path(op, (char *)buffer);
        remove_disk(op);
      }
#if EXTRA_COMMANDS
      *cur_dir = 0;
#endif
    }

    // check if drive A is empty
    if (drives[DRIVE_A].hnd == 0)
    {
      // if so, try insert disk "system.p32" from root directory of SD card
      strcpy_P((char *)buffer, PSTR("system.p32"));
      insert_disk(DRIVE_A, (char *)buffer, 0);
      config_write_pmd32_img_path(DRIVE_A, (char *)buffer);
      config_write_pmd32_wp(DRIVE_A, 0);
    }

    // turn on display backlight
    lcd_backlight_on();
    op = -99;

    while (1)
    {
      if (!sd_raw_available())  // card removed "on the fly" !!!
        break;

      if (op == -99)
      {
#if EXTRA_COMMANDS
        if (*cur_dir == '\0')
          fat_get_dir_entry_of_path(fs, "/", &de);
        else
          fat_get_dir_entry_of_path(fs, cur_dir, &de);
        ds = fat_open_dir(fs, &de);
#endif

        lcd_clr();
        lcd_disp_string(apputil_center_text_p(str, pstr_pmd32_ready, LCD_DISPLAY_COLUMNS));
        if (sd_raw_locked())
          lcd_disp_string_at_p(pstr_err_card_locked, 1, 0);
        op = 0;
      }

      if (btn_test() != 0) // test buttons
      {
#if DEBUG_ON_LCD_2
        lcd_disp_char_at(' ', 1, 0);
#endif
        val = btn_read();
        if (val == BTN_CANCEL)
          break;
        else if (val == BTN_SET)
        {
          sd_raw_sync();
#if EXTRA_COMMANDS
          if (ds)
          {
            fat_close_dir(ds);
            ds = 0;
          }
#endif
          settings();
          op = -99;
        }
        continue;
      }

      if (!send_byte(PRESENTATION, SEND_PRESENT_TIMEOUT))
        continue;
#if DEBUG_ON_LCD_2
        lcd_disp_char_at('s', 1, 0);
#endif
      if (!read_byte(&val, WAIT_PRESENT_TIMEOUT)) // wait presentation
        continue;
#if DEBUG_ON_LCD_2
        lcd_disp_char_at('r', 1, 0);
#endif
      if (val != PRESENTATION)
        continue;
#if DEBUG_ON_LCD_2
      lcd_disp_char_at('P', 1, 0);
#endif
#if DEBUG_CMD
      uart_puts_p(PSTR("Present\n"));
#endif

      while (1)
      {
        if (!read_byte(&command, WAIT_COMMAND_TIMEOUT)) // wait command
          break;
        if (command == PRESENTATION)
          continue;
        crc = command;
#if DEBUG_ON_LCD_2
        lcd_disp_char_at(command, 1, 0);
#endif
#if DEBUG_CMD
        uart_puts_p(PSTR("Cmd:"));
        uart_putc(command);
        uart_putc('\n');
#endif

        op = OP_NO_TRANSFER;
        switch (command)
        {
          case 'B' :  // BOOT
          case '*' :  // FAST mode - not implemented
          case '@' :  // SLOW mode - not implemented
#if EXTRA_COMMANDS
          case 'K' :  // get current path
#endif
            if (read_byte_crc_check())
            {
              if (command == 'B')
              {
                drvnum = 0;
                track = 0;
                sector = 0;
                op = OP_SEND_DATA; // send data
                length = SECTOR_SIZE;
              }
#if EXTRA_COMMANDS
              else if (command == 'K')
              {
                op = OP_SEND_DATA; // send data
                length = strlen(cur_dir);
              }
#endif
            }
            else
              op = OP_ERROR;
            break;

          case 'Q' :  // read logical sector
          case 'R' :  // read logical sector
          case 'F' :  // format track
            if (read_sector_track() && read_byte_crc_check())
            {
              if (command == 'F')
                op = OP_FORMAT; // format
              else
              {
                op = OP_SEND_DATA; // send data
                length = SECTOR_SIZE;
              }
            }
            else
              op = OP_ERROR;
            break;

          case 'T' :  // write logical sector
          case 'W' :  // write logical sector
          case 'S' :  // write physical sector
            if (read_sector_track())
            {
              op = OP_RECEIVE_DATA; // receive data
              if (command == 'S')
                length = PHYSICAL_SECTOR_SIZE + 1;
              else
                length = SECTOR_SIZE;
            }
            else
              op = OP_ERROR;
            break;

          case 'I' :  // select drive + HOME
#if EXTRA_COMMANDS
          case 'G' :  // get path/name of inserted image
          case 'L' :  // catalogue list
#endif
            op = OP_ERROR;
            if (read_byte_ex(&val))
            {
#if EXTRA_COMMANDS
              if (command == 'L')
              {
                if (val == 0)
                  fat_reset_dir(ds);
              }
              else
#endif
                drvnum = val;
              if (read_byte_crc_check())
              {
#if EXTRA_COMMANDS
                if (command != 'I')
                  op = OP_SEND_DATA;
                else
#endif
                  op = OP_NO_TRANSFER;
              }
            }
            break;

#if EXTRA_COMMANDS
          case 'H' :  // insert/remove image
            op = OP_ERROR;
            if (!read_byte_ex(&val))
              break;
            drvnum = val;
            if (!read_byte_ex(&val))
              break;
            sector = val; // write protect
            // fall through next case

          case 'M' :  // change directory
          case 'N' :  // create image
            op = OP_ERROR;
            if (read_byte_ex(&val))
            {
              length = val;
              op = OP_RECEIVE_DATA; // receive data
            }
            break;
#endif

          case 'U' :  // write to PMD 32 RAM - not implemented
          case 'C' :  // read from PMD 32 RAM - not implemented
          case 'J' :  // execute routine in PMD 32 RAM - not implemented
            op = OP_ERROR;
            if (read_byte_ex(&val))
            {
              address = (val << 8);
              if (read_byte_ex(&val))
              {
                address |= val;
                if (command == 'J')
                {
                  if (read_byte_crc_check())
                    op = OP_NO_TRANSFER;
                }
                else
                {
                  if (read_byte_ex(&val))
                  {
                    length = (val << 8);
                    if (read_byte_ex(&val))
                    {
                      length |= val;
                      if (command == 'C')
                      {
                        if (read_byte_crc_check())
                          op = OP_SEND_DATA; // send data
                      }
                      else
                        op = OP_RECEIVE_DATA; // receive data
                    }
                  }
                }
              }
            }
            break;

          default :
            op = OP_ERROR;
            break;
        }

        if (op == OP_NO_TRANSFER)   // ** WITHOUT ANY DATA TRANSFER **
        {
          if (command == 'I' && drvnum >= DRIVES_COUNT)
          {
            send_byte(RESULT_BD, SEND_TIMEOUT); // bad drive number
            op = OP_ERROR;
          }
          else if (!send_byte(RESULT_OK, SEND_TIMEOUT)) // result OK
          {
#if DEBUG_OP
            uart_puts_p(PSTR("op0:1\n"));
#endif
            op = OP_ERROR;
          }
        }
        else if (op == OP_SEND_DATA)  // ** SEND DATA **
        {
#if EXTRA_COMMANDS
          if (command == 'G')
          {
            if (drvnum >= DRIVES_COUNT)
            {
              send_byte(RESULT_BD, SEND_TIMEOUT); // bad drive number
              op = OP_ERROR;
            }
            else
            {
              memset(buffer, 0, sizeof(buffer));
//              *buffer = 0; // result OK
              config_read_pmd32_wp(drvnum, (buffer + 1)); // WP
              config_read_pmd32_img_path(drvnum, (char *)(buffer + 3)); // path to image
              length = strlen((char *)(buffer + 3));
              *(buffer + 2) = (uint8_t)length; // LEN
              length += 3;
            }
          }
          else if (command == 'K')
          {
            *buffer = (uint8_t)length;
            if (length)
              strcpy((char *)(buffer + 1), cur_dir);
            length++;
          }
          else if (command == 'L')
          {
            *buffer = 0;
            length = 1;
            while (fat_read_dir(ds, &de))
            {
              if (de.attributes & FAT_ATTRIB_VOLUME)
                continue;
              else if (de.attributes & FAT_ATTRIB_DIR)
              {
                if (strcmp(de.long_name, ".") == 0)
                  continue;
              }
              point = buffer + 1;
              if (de.attributes & FAT_ATTRIB_DIR)
                *point++ = '[';
              strcpy((char *)point, de.long_name);
              point += strlen(de.long_name);
              if (de.attributes & FAT_ATTRIB_DIR)
                *point++ = ']';
              length = point - buffer;
              *buffer = (uint8_t)(length - 1);
              break;
            }
          }
          else
#endif
          if (command != 'C')
          {
            if (!prepare_sector())
            {
#if DEBUG_OP
              uart_puts_p(PSTR("RdErr\n"));
#endif
              send_byte(RESULT_RE, SEND_TIMEOUT); // read sector error
              op = OP_ERROR;
            }
            else if (!send_byte(RESULT_OK, SEND_TIMEOUT)) // read sector OK
            {
#if DEBUG_OP
              uart_puts_p(PSTR("op1:1\n"));
#endif
              op = OP_ERROR;
            }
          }

          if (op != OP_ERROR)
          {
            point = buffer;
            crc = 0;
            while (length)
            {
              if (command == 'C')
                val = 0xAA;
              else
                val = *point++;
              crc ^= val;
              if (!send_byte(val, SEND_TIMEOUT))
              {
#if DEBUG_OP
                uart_puts_p(PSTR("op1:2\n"));
#endif
                op = OP_ERROR;
                break;
              }
              length--;
            }

            if (op != OP_ERROR)
            {
              if (!send_byte(crc, SEND_TIMEOUT))
              {
#if DEBUG_OP
                uart_puts_p(PSTR("op1:3\n"));
#endif
                op = OP_ERROR;
              }
            }
          }
        }
        else if (op == OP_RECEIVE_DATA || op == OP_FORMAT) // ** RECEIVE DATA / FORMAT **
        {
          point = buffer;
          if (op == OP_RECEIVE_DATA)  // ** receive data **
          {
            if (command == 'H' && length == 0xFF)
            {
              *point++ = 0xFF;
              length = 0;
            }

            while (length--)
            {
              if (!read_byte_ex(&val))
              {
#if DEBUG_OP
                uart_puts_p(PSTR("op2:1\n"));
#endif
                op = OP_ERROR;
                break;
              }
              if (command != 'U')
                *point++ = val;
            }
            if (op != OP_ERROR)
            {
              if (!read_byte_crc_check())
              {
#if DEBUG_OP
                uart_puts_p(PSTR("op2:2\n"));
#endif
                op = OP_ERROR;
              }
            }
          }

          if (op != OP_ERROR)
          {
#if EXTRA_COMMANDS
            if (command == 'M')
            {
              *point = 0;
              val = RESULT_NF;

              fat_reset_dir(ds);
              while (fat_read_dir(ds, &de))
              {
                if (de.attributes & FAT_ATTRIB_VOLUME)
                  continue;
                if ((de.attributes & FAT_ATTRIB_DIR) == 0)
                  continue;
                if (strcasecmp(de.long_name, (char *)buffer) == 0)
                {
                  if (strcmp(de.long_name, "..") == 0)
                  {
                    p = strrchr(cur_dir, '/');
                    if (p == 0)
                      *cur_dir = 0;
                    else
                      *p = 0;
                  }
                  else if (strcmp(de.long_name, ".") != 0)
                  {
                    if ((strlen(cur_dir) + 1 + strlen(de.long_name)) >= FAT_MAX_PATH)
                      val = RESULT_TL;
                    else
                    {
                      if (*cur_dir)
                        strcat(cur_dir, "/");
                      strcat(cur_dir, de.long_name);
                    }
                  }

                  if (val != RESULT_TL)
                  {
                    ds_new = fat_open_dir(fs, &de);
                    if (ds_new)
                    {
                      fat_close_dir(ds);
                      ds = ds_new;
                      val = RESULT_OK;

                    }
                  }
                  break;
                }
              }
            }
            else if (command == 'H' || command == 'N')
            {
              *point = 0;
              if (!(command == 'H' && *buffer == 0xFF))
              {
                length = strlen((char *)buffer);
                if (length > 3)
                {
                  if (strcasecmp_P((char *)(buffer + length - 3), pstr_p32_ext) != 0)
                  {
                    strcat((char *)buffer, ".");
                    strcat_P((char *)buffer, pstr_p32_ext);
                  }
                }
              }
              if (command == 'H' && drvnum >= DRIVES_COUNT)
                val = RESULT_BD;
              else if (*buffer == 0xFF)
              {
                drives[drvnum].wp = sector; // write protect
                config_write_pmd32_wp(drvnum, sector);
                val = RESULT_OK;
              }
              else if (length == 0)
              {
                if (command == 'H')
                {
                  remove_disk(drvnum);
                  memset(buffer, 0, FAT_MAX_PATH);
                  config_write_pmd32_img_path(drvnum, (char *)buffer);
                  val = RESULT_OK;
                }
                else
                  val = RESULT_CE;
              }
              else if ((strlen(cur_dir) + 1 + length) >= FAT_MAX_PATH)
                val = RESULT_TL;
              else
              {
                if (command == 'H')
                {
                  point = buffer + 128;
                  *point = 0;
                  if (*cur_dir)
                  {
                    strcpy((char *)point, cur_dir);
                    strcat((char *)point, "/");
                  }
                  strcat((char *)point, (char *)buffer);
                  insert_disk(drvnum, (char *)point, sector);
                  if (drives[drvnum].hnd == 0)
                  {
                    memset(point, 0, FAT_MAX_PATH);
                    config_write_pmd32_img_path(drvnum, (char *)point);
                    val = RESULT_NF;
                  }
                  else
                  {
                    config_write_pmd32_img_path(drvnum, (char *)point);
                    config_write_pmd32_wp(drvnum, sector);
                    val = RESULT_OK;
                  }
                }
                else  // 'N'
                {
                  if (create_image(ds, (const char *)buffer))
                    val = RESULT_OK;
                  else
                    val = RESULT_CE;
                }
              }
            }
            else
#endif
            if (command != 'U' && (drives[drvnum].wp || sd_raw_locked()))
              val = RESULT_WP;
            else
            {
              if (command == 'U' || write_sector())
              {
                if (command == 'F')
                  _delay_ms(300);
                val = RESULT_OK;
              }
              else
              {
                if (command == 'F')
                  val = RESULT_FE;
                else
                  val = RESULT_WE;
              }
            }
#if DEBUG_CMD
            uart_puts_p(PSTR("Res="));
            uart_putc_hex(val);
            uart_putc('\n');
#endif
            if (!send_byte(val, SEND_TIMEOUT))
            {
#if DEBUG_OP
              uart_puts_p(PSTR("op2:3\n"));
#endif
              op = OP_ERROR;
            }
          }
        }

        if (op == OP_ERROR)   // error
        {
#if DEBUG_OP
          uart_puts_p(PSTR("Err\n"));
#endif
          break;
        }
      }
    }

#if EXTRA_COMMANDS
    if (ds)
    {
      fat_close_dir(ds);
      ds = 0;
    }
#endif
    for (op = 0; op < DRIVES_COUNT; op++)
      remove_disk(op);

    close_sd_card();

    lcd_disp_string_at(apputil_center_text_p(str, pstr_pmd32_remove_0, LCD_DISPLAY_COLUMNS), 0, 0);
    lcd_disp_string_at(apputil_center_text_p(str, pstr_pmd32_remove_1, LCD_DISPLAY_COLUMNS), 1, 0);
    while (sd_raw_available())
    {
      if (btn_test())
      {
        if (btn_read() == BTN_SET)
          break;
      }
    }
  }
}
//----------------------------------------------------------------------------
uint8_t read_byte(uint8_t *byte, uint32_t to_ms)
{
  uint32_t next_ticks = timer_get_ticks() + to_ms;
  while (timer_get_ticks() < next_ticks)
  {
    if (pmd32_byte_available())
    {
      // switch MH8286 driver to output for reading from computer
      pmd32_dir_input();
      // ACK to low
      pmd32_ack_on();
      _delay_us(1);
      // read data
      *byte = pmd32_data_read();
      // ACK to high
      pmd32_ack_off();
      return 1;
    }
  }

  return 0;
}
//----------------------------------------------------------------------------
uint8_t read_byte_ex(uint8_t *byte)
{
  if (read_byte(byte, WAIT_BYTE_TIMEOUT))
  {
    crc ^= *byte;
    return 1;
  }

  send_byte(NAK, SEND_NAK_TIMEOUT);
  return 0;
}
//----------------------------------------------------------------------------
uint8_t read_byte_crc_check(void)
{
  uint8_t b;

  if (read_byte(&b, WAIT_BYTE_TIMEOUT))
  {
#if DEBUG_CRC
    uart_puts_p(PSTR("CRC:"));
    uart_putc_hex(crc);
    uart_putc(',');
    uart_putc_hex(b);
    uart_putc('\n');
#endif
    crc ^= b;
    if (!crc)
    {
      if (send_byte(ACK, SEND_ACK_TIMEOUT))
      {
#if DEBUG_CRC
        uart_puts_p(PSTR("CRC ok\n"));
#endif
        return 1;
      }
    }
  }

#if DEBUG_CRC
  uart_puts_p(PSTR("CRC err\n"));
#endif
  send_byte(NAK, SEND_NAK_TIMEOUT);
  return 0;
}
//----------------------------------------------------------------------------
uint8_t read_sector_track(void)
{
  uint8_t b;

  // drive + sector
  if (!read_byte_ex(&b))
    return 0;
  drvnum = (b >> 6);
  if (drvnum == 1 || drvnum == 2)
    drvnum ^= 3;
  sector = b & 0x3F;
#if DEBUG_CMD
  uart_puts_p(PSTR("Drv:"));
  uart_putc('A' + drvnum);
  uart_puts_p(PSTR("\nSec:"));
  uart_putc_hex(sector);
  uart_putc('\n');
#endif

  // track
  if (!read_byte_ex(&b))
    return 0;
  track = b;
#if DEBUG_CMD
  uart_puts_p(PSTR("Trk:"));
  uart_putc_hex(track);
  uart_putc('\n');
#endif

  return 1;
}
//----------------------------------------------------------------------------
uint8_t send_byte(uint8_t byte, uint32_t to_ms)
{
  // switch MH8286 driver to input for writing to computer
  pmd32_dir_output();
  // write data
  pmd32_data_write(byte);
  // port as output
  pmd32_data_port_out();
  // STB to low
  pmd32_stb_on();
  _delay_us(1);
  // STB to high
  pmd32_stb_off();
  _delay_us(1);

  // data port back to Z-state
  pmd32_data_port_Z();

  if (to_ms == 0)
    return 1;

  uint32_t next_ticks = timer_get_ticks() + to_ms;
  while (timer_get_ticks() < next_ticks)
  {
    if (pmd32_byte_accepted())
      return 1;
  }

  return 0;
}
//----------------------------------------------------------------------------
uint8_t prepare_sector(void)
{
  if (drives[drvnum].hnd != 0 && track < TRACKS && sector < SECTORS_PER_TRACK)
  {
    int32_t seek = (int32_t)track * (int32_t)TRACK_SIZE + (int32_t)sector * (int32_t)SECTOR_SIZE;
#if DEBUG_SD
    uart_puts_p(PSTR("Seek:"));
    uart_putdw_hex(seek);
    uart_putc('\n');
#endif

    if (fat_seek_file(drives[drvnum].hnd, &seek, FAT_SEEK_SET))
    {
      if (fat_read_file(drives[drvnum].hnd, buffer, SECTOR_SIZE) == SECTOR_SIZE)
        return 1;
    }
  }

  return 0;
}
//----------------------------------------------------------------------------
uint8_t write_sector(void)
{
  uint32_t size;

  if (command == 'S')
  {
    sector &= 0x3C;
    size = PHYSICAL_SECTOR_SIZE;
  }
  else if (command == 'T' || command == 'W')
    size = SECTOR_SIZE;
  else  // 'F'
  {
    size = TRACK_SIZE;
    sector = 0;
    memset(buffer, 0xE5, sizeof(buffer));
  }

  if (drives[drvnum].hnd != 0 && track < TRACKS && sector < SECTORS_PER_TRACK)
  {
    int32_t seek = (int32_t)track * (int32_t)TRACK_SIZE + (int32_t)sector * (int32_t)SECTOR_SIZE;
#if DEBUG_SD
    uart_puts_p(PSTR("Seek:"));
    uart_putdw_hex(seek);
    uart_putc('\n');
#endif

    if (fat_seek_file(drives[drvnum].hnd, &seek, FAT_SEEK_SET))
    {
      while (size >= PHYSICAL_SECTOR_SIZE)
      {
        if (fat_write_file(drives[drvnum].hnd, buffer, PHYSICAL_SECTOR_SIZE) != PHYSICAL_SECTOR_SIZE)
        {
#if DEBUG_SD
          uart_puts_p(PSTR("Write error 1\n"));
#endif
          return 0;
        }
        size -= PHYSICAL_SECTOR_SIZE;
      }

      if (size > 0)
      {
        if (fat_write_file(drives[drvnum].hnd, buffer, size) == size)
          return 1;
#if DEBUG_SD
        else
        {
          uart_puts_p(PSTR("Write error 2\n"));
        }
#endif
      }
      else
        return 1;
    }
#if DEBUG_SD
    else
    {
      uart_puts_p(PSTR("Seek error\n"));
    }
#endif
  }

  return 0;
}
//---------------------------------------------------------------------------
void insert_disk(uint8_t drive, char *file_path, uint8_t wp)
{
  struct fat_dir_entry_struct dir_entry;

  if (drive >= DRIVES_COUNT)
    return;

  remove_disk(drive);

  if (fat_get_dir_entry_of_path(fs, file_path, &dir_entry))
  {
    drives[drive].hnd = fat_open_file(fs, &dir_entry);
    drives[drive].wp = wp;
  }
}
//---------------------------------------------------------------------------
void remove_disk(uint8_t drive)
{
  if (drive >= DRIVES_COUNT)
    return;

  if (drives[drive].hnd != 0)
  {
    fat_close_file(drives[drive].hnd);
    drives[drive].hnd = 0;
  }
}
//----------------------------------------------------------------------------
uint8_t create_image(struct fat_dir_struct *parent_dir, const char *file_name)
{
  struct fat_dir_entry_struct dir_entry;
  struct fat_file_struct *file;
  int32_t total;
  uint8_t rv;

  if (!fat_create_file(parent_dir, file_name, &dir_entry))
  {
#if DEBUG_ON_LCD
    lcd_disp_char_at('0', 0, 0);
    btn_read();
#endif
#if DEBUG_SD
    uart_puts_p(PSTR("fat_create_file error\n"));
#endif
    return 0;
  }

  file = fat_open_file(fs, &dir_entry);
  if (!file)
  {
#if DEBUG_ON_LCD
    lcd_disp_char_at('1', 0, 0);
    btn_read();
#endif
#if DEBUG_SD
    uart_puts_p(PSTR("fat_open_file error\n"));
#endif
    return 0;
  }

  rv = 0;
  total = 0;
  if (fat_seek_file(file, &total, FAT_SEEK_SET))
  {
    memset(buffer, 0xE5, sizeof(buffer));
    while (total < DISK_SIZE)
    {
      if (fat_write_file(file, buffer, PHYSICAL_SECTOR_SIZE) != PHYSICAL_SECTOR_SIZE)
      {
#if DEBUG_ON_LCD
        lcd_disp_char_at('2', 0, 0);
        btn_read();
#endif
        break;
      }

      total += PHYSICAL_SECTOR_SIZE;
    }
  }
#if DEBUG_ON_LCD
  else
  {
    lcd_disp_char_at('2', 0, 0);
    btn_read();
  }
#endif
  if (total == DISK_SIZE)
    rv = 1;

#if DEBUG_SD
    uart_putdw_dec((uint32_t)total);
    uart_putc('\n');
#endif

  fat_close_file(file);

  return rv;
}
//----------------------------------------------------------------------------
#define FN_LEN          12
#define EXT_LEN         3
//----------------------------------------------------------------------------
#define MI_DRIVE_A      0
#define MI_DRIVE_B      1
#define MI_DRIVE_C      2
#define MI_DRIVE_D      3
#define MI_CREATE_IMG   4
//----------------------------------------------------------------------------
void settings(void)
{
  int8_t mi = 0, btn = BTN_CANCEL, wp;
  char path[FAT_MAX_PATH];
  char fn[FN_LEN + EXT_LEN + 1];

  // PMD 32 menu
  MENU_ITEM pmd32_menu_items[] = {
    { pstr_pmd32_menu_0, MI_DRIVE_A },
    { pstr_pmd32_menu_1, MI_DRIVE_B },
    { pstr_pmd32_menu_2, MI_DRIVE_C },
    { pstr_pmd32_menu_3, MI_DRIVE_D },
    { pstr_pmd32_menu_4, MI_CREATE_IMG }
  };
  MENU pmd32_menu = {
    pstr_pmd32_menu_title,
    sizeof(pmd32_menu_items) / sizeof(MENU_ITEM),
    pmd32_menu_items
  };

  while (1)
  {
    mi = ui_menu(&pmd32_menu, mi);
    if (mi < 0)
      return;

    if (mi == MI_CREATE_IMG) // create new image
    {
      if (sd_raw_locked())
      {
        ui_show_message(pstr_err_card_locked, pstr_any_button);
        continue;
      }
      
      if (!ui_select_file(pstr_select_directory, fs, 0,
                          path, FAT_MAX_PATH, 1))
        continue;
      if (!ui_input_string(pstr_enter_image_name, fn, FN_LEN))
        continue;
      strcat(fn, ".");
      strcat_P(fn, pstr_p32_ext);

      if (strlen(path) + 1 + strlen(fn) >= FAT_MAX_PATH)
      {
        ui_show_message(pstr_path_too_long, pstr_any_button);
      }
      else
      {
        ui_wait_please(pmd32_menu_items[mi].name);

        struct fat_dir_entry_struct de;
        struct fat_dir_struct* dd;
        fat_get_dir_entry_of_path(fs, path, &de);
        dd = fat_open_dir(fs, &de);
        if (!create_image(dd, fn))
          ui_show_message(pstr_err_create_file, pstr_any_button);
        if (dd)
          fat_close_dir(dd);
      }
    }
    else // manage drives
    {
      wp = drives[mi].wp; // remember old WP setting
      while (1)
      {
        if (btn != BTN_SET)
        {
          memset(path, ' ', sizeof(path));
          path[0] = (char)('A' + mi);
          path[1] = ':';
          if (drives[mi].hnd)
            apputil_extract_file_name(path + 2, LCD_DISPLAY_COLUMNS - 5, (drives[mi].hnd)->dir_entry.long_name);
          else
            strcpy_P(path + 2, pstr_empty);

          lcd_clr();
          lcd_disp_string_at(path, 0, 0);
          lcd_disp_string_at_p(pstr_manage_drive, 1, 0);
        }

        if (drives[mi].hnd)
        {
          lcd_disp_char_at('R', 0, LCD_DISPLAY_COLUMNS - 2);
          if (wp)
            lcd_disp_char('O');
          else
            lcd_disp_char('W');
        }

        btn = btn_read();
        if (btn == BTN_RIGHT) // insert image
        {
          if (ui_select_file(pstr_select_p32_image, fs, pstr_p32_ext,
                             path, FAT_MAX_PATH, 0))
          {
            ui_wait_please(pmd32_menu_items[mi].name);
            config_write_pmd32_img_path(mi, path);
            insert_disk(mi, path, 1);
            wp = 1;
          }
        }
        else if (btn == BTN_LEFT) // remove image
        {
          ui_wait_please(pmd32_menu_items[mi].name);
          memset(path, 0, FAT_MAX_PATH);
          config_write_pmd32_img_path(mi, path);
          remove_disk(mi);
        }
        else if (btn == BTN_SET) // change write protect state
        {
          wp ^= 1;
        }
        else if (btn == BTN_CANCEL) // escape drive settings
        {
          if (wp != drives[mi].wp) {
            drives[mi].wp = wp;
            config_write_pmd32_wp(mi, wp);
          }
          break;
        }
      }
    }
  }
}
//----------------------------------------------------------------------------
void close_sd_card(void)
{
  if (part && fs)
    sd_raw_sync();

  // close file system
  if (fs)
  {
    fat_close(fs);
    fs = 0;
  }

  // close partition first
  if (part)
  {
    partition_close(part);
    part = 0;
  }
}
//----------------------------------------------------------------------------
void prepare_sd_card(void)
{
  PGM_P ptxt = 0;

  close_sd_card();

  while (1)
  {
    if (ptxt)
    {
      lcd_disp_string_at(apputil_center_text_p(str, ptxt, LCD_DISPLAY_COLUMNS), 0, 0);
      lcd_disp_string_at_p(pstr_any_button, 1, 0);
      btn_read();
      ptxt = 0;
    }

    if (!sd_raw_init())
    {
      if (!sd_raw_available())
        ptxt = pstr_err_insert_card;
      else
        ptxt = pstr_err_init_card;
      continue;
    }

    // open first partition
    part = partition_open(sd_raw_read, sd_raw_read_interval,
                          sd_raw_write, sd_raw_write_interval, 0);
    if (!part)
    {
      /* If the partition did not open, assume the storage device
       * is a "superfloppy", i.e. has no MBR.
       */
      part = partition_open(sd_raw_read, sd_raw_read_interval,
                            sd_raw_write, sd_raw_write_interval, -1);
      if (!part)
      {
        ptxt = pstr_err_partition;
        continue;
      }
    }

    // open file system
    fs = fat_open(part);
    if (!fs)
    {
      ptxt = pstr_err_fat;
      continue;
    }

    break;
  }
}
//----------------------------------------------------------------------------
